# numpy examples --------------------------------------------------------------

# import
import numpy as np


# probability -----------------------------------------------------------------


def estimate_probability(n_sims=100):

    # one simulation of five random variables per row
    X = np.random.normal(size=(n_sims, 5))

    # get maximum magnitude for each simulation / row
    row_max = np.max(np.abs(X), axis=1)

    # get proportion of rows that meet the condition / estimate probability
    probability = np.mean(row_max > 2)

    return probability


# example usage
estimate_probability(10)
estimate_probability(100)
estimate_probability(1000)
estimate_probability(10000)

# checkerboard ----------------------------------------------------------------


def make_checkerboard(n, top_left=0):

    # create an nxn board of 0s
    board = np.zeros((n, n), dtype=int)

    # cleverly place 1s
    board[::2, 1::2] = 1
    board[1::2, ::2] = 1

    # flip board if requested
    if top_left == 1:
        # board = np.abs(board - 1)
        # board = board[:, ::-1] # only works for even n
        board = board ^ top_left

    return board


print(make_checkerboard(n=10))
print(make_checkerboard(n=6, top_left=1))


# series ----------------------------------------------------------------------

## geometric series -----------------------------------------------------------


def make_geometric_series(a, r, n_terms, verbose=False):
    if not isinstance(a, (int, float)) or not isinstance(r, (int, float)):
        raise ValueError("Both 'a' and 'r' must be real numbers.")
    if not isinstance(n_terms, int) or n_terms < 1:
        raise ValueError("'n_terms' must be an integer greater than or equal to 1.")
    if not abs(r) < 1:
        raise ValueError("The ratio 'r' must have a magnitude less than 1.")

    # create terms of series
    terms = a * r ** np.arange(n_terms)

    # sum terms to get numerical approximation
    numeric = np.sum(terms)

    # calculate analytic solution
    true = a / (1 - r)

    # print details if requested
    if verbose:
        print(f"The numerical approximation is: {numeric}")
        print(f"The true value is:              {true}")
        if np.allclose(numeric, true):
            print("Converged!")
        else:
            print("Buddy, you need more terms!")

    # return approximation, true value, and calculated terms
    return {"numeric": numeric, "true": true, "terms": terms}


# example usage
print(make_geometric_series(5, 0.3, 6)["terms"])
_ = make_geometric_series(5, 0.3, 2, verbose=True)
_ = make_geometric_series(2, 0.8, 10, verbose=True)


## geometric series -----------------------------------------------------------


def make_exponential_series(x, n_terms, verbose=False):
    if not isinstance(x, (int, float)):
        raise ValueError("`x` must be a real number.")
    if not isinstance(n_terms, int) or n_terms < 1:
        raise ValueError("'n_terms' must be an integer greater than or equal to 1.")

    # create numerator for each term
    numerators = x ** np.arange(n_terms)

    # create denominator for each term
    factorial_inputs = np.arange(0, n_terms)
    factorial_inputs[0] = 1
    denominators = np.cumprod(factorial_inputs)

    # create terms of the series
    terms = numerators / denominators

    # sum terms to get numerical approximation
    numeric = np.sum(terms)

    # calculate analytic solution
    true = np.exp(x)

    # print details if requested
    if verbose:
        print(f"The numerical approximation is: {numeric}")
        print(f"The true value is:              {true}")
        if np.allclose(numeric, true):
            print("Converged!")
        else:
            print("Buddy, you need more terms!")

    # return approximation, true value, and calculated terms
    return {"numeric": numeric, "true": true, "terms": terms}


# example usage
print(make_exponential_series(3, n_terms=5)["terms"])
_ = make_exponential_series(3, n_terms=5, verbose=True)
_ = make_exponential_series(0.5, n_terms=50, verbose=True)


# knn -------------------------------------------------------------------------


# create knn class
class KnnRegressor:
    def __init__(self, k=5):
        self.k = k

    def fit(self, X, y):
        self.X = X
        self.y = y

    def __repr__(self) -> str:
        return f"KNN Regression model with k = {self.k}."

    def predict(self, X_new):
        predicted_labels = [self._predict(x) for x in X_new]
        return np.array(predicted_labels)

    def _predict(self, x_new):
        # compute distances between new x and all samples in the X data
        distances = [np.linalg.norm(x_new - x) for x in self.X]
        # sort by distance and return indices of the first k neighbors
        k_indices = np.argsort(distances)[: self.k]
        # extract the labels of the k nearest neighbor training samples
        k_nearest_y = self.y[k_indices]
        # return the mean of the k nearest neighbors
        return np.mean(k_nearest_y)


# import sklearn version of knn
from sklearn.neighbors import KNeighborsRegressor

# create a 50x2 matrix X with random values
X = np.random.rand(25, 2)

# create a target variable y as a linear combination of columns of X plus some noise
y = X[:, 0] + 2 * X[:, 1] + np.random.rand(25)

# fit and predict using numpy only knn
knn_np = KnnRegressor(k=5)
knn_np.fit(X, y)
knn_np_pred = knn_np.predict(X)

# fit and predict using knn from sklearn
knn_sk = KNeighborsRegressor(n_neighbors=5)
_ = knn_sk.fit(X, y)
knn_sk_pred = knn_sk.predict(X)

# compare predictions
print(knn_np_pred)
print(knn_sk_pred)
print(all(knn_np_pred == knn_sk_pred))
